@font-face{font-family:CustomFont;src:url('../fonts/custom-font.eot') format('embedded-opentype'),url('../fonts/custom-font.woff') format('woff'),url('../fonts/custom-font.ttf') format('truetype'),url('../fonts/custom-font.svg') format('svg');font-weight:400;font-style:normal}.cf{display:inline-block;font:normal normal normal 14px/1 CustomFont;font-size:inherit;text-rendering:auto;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}.cf-scatter-chart:before{content:"\e800"}